/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_GTKUI

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <sys/types.h> /* size_t used by regex */
#include <regex.h>     /* regcomp, regexec, regfree */
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <libetpan.h>

#include "intl.h"


#include "x_alloc.h"
#include "gtk-utils.h"
#include "logging.h"
#include "mailbox.h"
#include "folder.h"
#include "manager.h"

#include "gtk-mb-editor.h"
#include "gtk-server-editor.h"
#include "gtkxtreeview.h"

typedef struct XePGtkMbEditorFolder_s XePGtkMbEditorFolder;

struct XePGtkMbEditorFolder_s
{
  char *path;
  int flags;
};

enum
  {
    XEP_MBEDITOR_SUBSCRIBED = 1 << 0,
    XEP_MBEDITOR_HIDDEN     = 1 << 1,
    XEP_MBEDITOR_SELECTED   = 1 << 2
  };

enum
  {
    XEP_MBEDITOR_FOLDER_COL = 1,
    XEP_MBEDITOR_SUBSCRIBED_COL
  };

struct XePGtkMbEditor_s
{
  GtkWidget *widget;
  GtkListStore *liststore;
  GtkWidget *treeview;
  GtkWidget *notebook;
  XePMailbox *mailbox;
  int new_mailbox;
  void (*user_cb)(XePGtkMbEditor *ambeditor, void *auser_data);
  void *user_data;
  chash *folder_hash;
  clist *folder_list;

  int block;

  /* General */
  GtkWidget *name;
  GtkWidget *type_menu;

  /* Location */
  XePGtkServerEditor *seditor;

  GtkWidget *path_table;
  GtkWidget *path_entry;

  /* Folders */
  GtkWidget *xtreeview;
  GtkWidget *folders_widget;
  GtkWidget *folder_button;
  GtkWidget *namespace_widget;
  GtkWidget *subscribed_button;
  GtkWidget *personal_ns_button;
  GtkWidget *users_ns_button;
  GtkWidget *shared_ns_button;
  GtkWidget *search_entry;
  GtkWidget *search_button;
  GtkWidget *selected_button;
  GtkWidget *selection_widget;

  regex_t preg;
  gboolean valid_preg;

  /* Performance */
  GtkWidget *use_cache;
  GtkWidget *cache_entry;
  GtkWidget *flags_entry;
  GtkWidget *connections_widget;
  GtkWidget *connection_button;
  GtkWidget *max_connections;
};

enum
  {
    XEP_MBEDITOR_LOCATION_PAGE = 0,
    XEP_MBEDITOR_FOLDERS_PAGE,
    XEP_MBEDITOR_OPTIONS_PAGE,
    XEP_MBEDITOR_PERFORMANCE_PAGE
  };

enum
  {
    XEP_MBEDITOR_SHOW_SERVER      = 1 << 0,
    XEP_MBEDITOR_SHOW_PATH        = 1 << 1,
    XEP_MBEDITOR_SHOW_APOP        = 1 << 2,
    XEP_MBEDITOR_SHOW_FOLDERS     = 1 << 3,
    XEP_MBEDITOR_SHOW_SUBSCRIBED  = 1 << 4,
    XEP_MBEDITOR_SHOW_NAMESPACE   = 1 << 5,
    XEP_MBEDITOR_SHOW_CONNECTIONS = 1 << 6
  };

#define XEP_MBEDITOR_SHOW_IMAP \
  (XEP_MBEDITOR_SHOW_SERVER | XEP_MBEDITOR_SHOW_FOLDERS | \
   XEP_MBEDITOR_SHOW_SUBSCRIBED | XEP_MBEDITOR_SHOW_NAMESPACE | \
   XEP_MBEDITOR_SHOW_CONNECTIONS)

#define XEP_MBEDITOR_SHOW_MBOX \
  (XEP_MBEDITOR_SHOW_PATH)

#define XEP_MBEDITOR_SHOW_MH \
  (XEP_MBEDITOR_SHOW_PATH | XEP_MBEDITOR_SHOW_FOLDERS | \
   XEP_MBEDITOR_SHOW_SUBSCRIBED)

#define XEP_MBEDITOR_SHOW_POP3 \
  (XEP_MBEDITOR_SHOW_SERVER | XEP_MBEDITOR_SHOW_APOP)

#define XEP_MBEDITOR_SHOW_NNTP \
  (XEP_MBEDITOR_SHOW_SERVER | XEP_MBEDITOR_SHOW_FOLDERS | \
   XEP_MBEDITOR_SHOW_SUBSCRIBED | XEP_MBEDITOR_SHOW_CONNECTIONS)

#define XEP_MBEDITOR_CHASH_MAXSIZE 10000

/* Static functions declarations */

static gboolean
GtkMbEditor_delete_event_cb(GtkWidget *widget, 
			    GdkEvent *event, 
			    gpointer ambeditor);

static void
GtkMbEditor_destroy_cb(GObject *object, gpointer ambeditor);

static void
GtkMbEditor_ok_cb(GtkButton *button, gpointer ambeditor);

static void
GtkMbEditor_cancel_cb(GtkButton *button, gpointer ambeditor);

static gboolean
GtkMbEditor_is_blocked(XePGtkMbEditor *mbeditor);

static void
GtkMbEditor_append_location_page(XePGtkMbEditor *mbeditor);

static void
GtkMbEditor_append_folders_page(XePGtkMbEditor *mbeditor);

static void
GtkMbEditor_append_options_page(XePGtkMbEditor *mbeditor);

static void
GtkMbEditor_append_performance_page(XePGtkMbEditor *mbeditor);

GtkWidget *
GtkMbEditor_type_menu_create(XePGtkMbEditor *mbeditor);

static void
GtkMbEditor_show_mailbox(XePGtkMbEditor *mbeditor);

static unsigned int
GtkMbEditor_get_widgets_from_type(XePMailboxType type);

static void
GtkMbEditor_set_type_cb(gpointer ambeditor, guint mboxtype, GtkWidget *widget);

static void
GtkMbEditor_apply_changes(XePGtkMbEditor *mbeditor);

static void
GtkMbEditor_folders_init(XePGtkMbEditor *mbeditor);

static gboolean
GtkMbEditor_folder_data_func(void *iterator, 
			     guint col_id, 
			     void **content,
			     gchar **attribute,
			     gpointer data);

static void
GtkMbEditor_folder_params_func(void *iterator, gboolean *has_children,
			       guint *depth, gpointer data);

static gboolean
GtkMbEditor_folder_iterator_func(void **iterator, 
				 gboolean next,
				 gboolean ignore_children,
				 guint sort_col_id,
				 GtkSortType sort_order,
				 gpointer ambeditor);
static gboolean
GtkMbEditor_folder_match_func(void *iterator, gpointer ambeditor);

static gboolean
GtkMbEditor_folder_search_key_press_cb(GtkEditable *editable, 
				       GdkEventKey *event,
				       gpointer ambeditor);
static void
GtkMbEditor_folder_search_changed_cb(GtkEditable *editable, 
				     gpointer ambeditor);

static void
GtkMbEditor_folder_search_next_cb(GtkWidget *widget, gpointer ambeditor);

static void
GtkMbEditor_folders_subscribed_only_cb(GtkButton *button, gpointer ambeditor);

static void
GtkMbEditor_folders_subscribe_cb(GtkButton *button, gpointer ambeditor);

static void
GtkMbEditor_folders_unsubscribe_cb(GtkButton *button, gpointer ambeditor);

static void
GtkMbEditor_folders_select_all_cb(GtkButton *button, gpointer ambeditor);

static void
GtkMbEditor_folders_invert_cb(GtkButton *button, gpointer ambeditor);


static void
GtkMbEditor_get_folder_list_cb(GtkButton *button, gpointer ambeditor);

static void
GtkMbEditor_folder_list_reset(XePGtkMbEditor *mbeditor);

static void
GtkMbEditor_folder_list_cb(void *ambeditor, void *amb_list);

static void
GtkMbEditor_folder_lsub_cb(void *ambeditor, void *amb_lsub);

static void
GtkMbEditor_folder_list_register(XePGtkMbEditor *mbeditor, 
				 clist *folder_list, 
				 int flag);

static void
GtkMbEditor_folder_finalize_cb(void *ambeditor, void *anull);

static void
GtkMbEditor_set_folder_list(XePGtkMbEditor *mbeditor);

/* Functions definitions */

XePGtkMbEditor *
a_GtkMbEditor_new(GtkWidget *toplevel)
{
  XePGtkMbEditor *mbeditor;
  GtkWidget *vbox;
  GtkWidget *hbuttonbox;
  GtkWidget *cancelbutton;
  GtkWidget *okbutton;

  A_LOGGING_TRACE;

  mbeditor = x_alloc(1, XePGtkMbEditor);
  if (mbeditor == NULL)
    {
      A_LOGGING_ERROR("malloc(): %s", strerror(errno));
      return NULL;
    }

  mbeditor->widget = a_GtkUtils_transient_new(toplevel,
					      _("XetPan Mailbox Editor"),
					      -1, -1);
  gtk_window_set_modal(GTK_WINDOW(mbeditor->widget), TRUE);

  vbox = gtk_vbox_new(FALSE, 6);
  gtk_container_add(GTK_CONTAINER(mbeditor->widget), vbox);

  /* Notebook */

  mbeditor->notebook = gtk_notebook_new();
  gtk_notebook_set_show_tabs(GTK_NOTEBOOK(mbeditor->notebook), TRUE);

  GtkMbEditor_append_location_page(mbeditor);
  GtkMbEditor_append_folders_page(mbeditor);
  GtkMbEditor_append_options_page(mbeditor);
  GtkMbEditor_append_performance_page(mbeditor);

  gtk_box_pack_start(GTK_BOX(vbox), mbeditor->notebook, TRUE, TRUE, 0);

  /* OK Cancel buttons */

  hbuttonbox = gtk_hbutton_box_new();
  gtk_button_box_set_layout(GTK_BUTTON_BOX(hbuttonbox), GTK_BUTTONBOX_END);
  gtk_box_set_spacing(GTK_BOX(hbuttonbox), 12);
  gtk_box_pack_start(GTK_BOX(vbox), hbuttonbox, FALSE, FALSE, 0);

  cancelbutton = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
  gtk_container_add(GTK_CONTAINER(hbuttonbox), cancelbutton);
  gtk_widget_show(cancelbutton);

  okbutton = gtk_button_new_from_stock(GTK_STOCK_OK);
  gtk_container_add(GTK_CONTAINER(hbuttonbox), okbutton);
  gtk_widget_show(okbutton);

  /* Show widgets */

  gtk_widget_show(hbuttonbox);
  gtk_widget_show(mbeditor->notebook);

  gtk_widget_show(vbox);

  /* Finish structure initialization */

  mbeditor->folder_hash = NULL;
  mbeditor->folder_list = NULL;
  mbeditor->block = 0;
  mbeditor->valid_preg = FALSE;

  /* Signals */

  g_signal_connect(G_OBJECT(mbeditor->widget), "delete-event",
		   G_CALLBACK(GtkMbEditor_delete_event_cb), mbeditor);

  g_signal_connect(G_OBJECT(mbeditor->widget), "destroy",
		   G_CALLBACK(GtkMbEditor_destroy_cb), mbeditor);

  g_signal_connect(G_OBJECT(okbutton), "clicked",
		   G_CALLBACK(GtkMbEditor_ok_cb), mbeditor);

  g_signal_connect(G_OBJECT(cancelbutton), "clicked",
		   G_CALLBACK(GtkMbEditor_cancel_cb), mbeditor);

  return mbeditor;
}

static gboolean
GtkMbEditor_delete_event_cb(GtkWidget *widget, 
			    GdkEvent *event, 
			    gpointer ambeditor)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;

  A_LOGGING_TRACE;

  if (GtkMbEditor_is_blocked(mbeditor))
    return TRUE;

  a_GtkMbEditor_hide(mbeditor);
  return TRUE;
}

static void
GtkMbEditor_destroy_cb(GObject *object, gpointer ambeditor)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;

  A_LOGGING_TRACE;

  /* FIXME: we should test if there are pending connections
   * (mbeditor->block non zero)
   */
  a_GtkServerEditor_free(mbeditor->seditor);
  GtkMbEditor_folder_list_reset(mbeditor);

  if (mbeditor->valid_preg)
    regfree(&mbeditor->preg);

  free(mbeditor);
}

static void
GtkMbEditor_ok_cb(GtkButton *button, gpointer ambeditor)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;
  XePMailbox *mailbox = mbeditor->mailbox;
  A_LOGGING_TRACE;

  if (GtkMbEditor_is_blocked(mbeditor))
    return;

  GtkMbEditor_apply_changes(mbeditor);

  if (mailbox->ep_mail_list)
    {
      mail_list_free(mailbox->ep_mail_list);
      mailbox->ep_mail_list = NULL;
    }

  if (mbeditor->mailbox->ep_mail_lsub)
    {
      mail_list_free(mailbox->ep_mail_lsub);
      mailbox->ep_mail_lsub = NULL;
    }

  if (mbeditor->user_cb)
    {
      mbeditor->user_cb(mbeditor, mbeditor->user_data);
    }

  a_GtkMbEditor_hide(mbeditor);
}

static void
GtkMbEditor_cancel_cb(GtkButton *button, gpointer ambeditor)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;

  A_LOGGING_TRACE;

  if (GtkMbEditor_is_blocked(mbeditor))
    return;

  if (mbeditor->new_mailbox)
    {
      a_Mailbox_free(mbeditor->mailbox);
    }
  a_GtkMbEditor_hide(mbeditor);
}

void
a_GtkMbEditor_show(XePGtkMbEditor *mbeditor)
{
  A_LOGGING_TRACE;

  gtk_notebook_set_current_page(GTK_NOTEBOOK(mbeditor->notebook), 0);
  gtk_widget_show(mbeditor->widget);
}

static gboolean
GtkMbEditor_is_blocked(XePGtkMbEditor *mbeditor)
{
  if (mbeditor->block)
    {
      a_GtkUtils_dialog_alert(_("Mailbox Editor Busy"),
			      mbeditor->widget,
			      XETPAN_WARNING_MESSAGE,
			      _("The mailbox editor is accessing the mailbox."),
			      _("Please wait until it has finished."),
			      XEP_GTKUTILS_OK,
			      TRUE, TRUE, NULL, NULL);
      return TRUE;
    }
  return FALSE;
}

void
a_GtkMbEditor_hide(XePGtkMbEditor *mbeditor)
{
  A_LOGGING_TRACE;

  gtk_widget_hide(mbeditor->widget);
}

void
a_GtkMbEditor_set_mailbox(XePGtkMbEditor *mbeditor, XePMailbox *mailbox)
{
  clist *list;
  clist *lsub;

  A_LOGGING_TRACE;
  
  GtkMbEditor_folder_list_reset(mbeditor);
  
  list = clist_new();
  lsub = clist_new();

  if (mailbox == NULL)
    {
      mailbox = a_Mailbox_new(_("(Enter new mailbox name)"));
      a_Mailbox_set_type(mailbox, XEP_MAILBOX_IMAP);
      mbeditor->mailbox = mailbox;      
      mbeditor->new_mailbox = 1;
    }
  else
    {
      clistiter *iter;
      XePFolder *folder;

      mbeditor->mailbox = mailbox;
      mbeditor->new_mailbox = 0;
      for (iter = clist_begin(mailbox->folders); iter; iter = iter->next)
	{
	  folder = (XePFolder *) iter->data;
	  clist_append(list, strdup(a_Folder_get_path(folder)));
	  if (folder->uid > 0)
	    {
	      /* folders with negative uid were unsubscribed by
	       * a previous instance of GtkMBEditor, so we 
	       * do not add them to the subscribed list
	       */
	      clist_append(lsub, strdup(a_Folder_get_path(folder)));
	    }
	}
    }
  mailbox->ep_mail_list = mail_list_new(list);
  mailbox->ep_mail_lsub = mail_list_new(lsub);
  GtkMbEditor_folder_list_register(mbeditor, list, 0);
  GtkMbEditor_folder_list_register(mbeditor, lsub, 
				   XEP_MBEDITOR_SUBSCRIBED);
  
  GtkMbEditor_show_mailbox(mbeditor);
}

static void
GtkMbEditor_show_mailbox(XePGtkMbEditor *mbeditor)
{
  unsigned int show;
  XePMailbox *mailbox = mbeditor->mailbox;
  XePGtkServerType server_type = XEP_SERVEREDITOR_NONE;
  int type_menu;

  A_LOGGING_TRACE;

  show = GtkMbEditor_get_widgets_from_type(mailbox->type);

  switch (mailbox->type)
    {
    case XEP_MAILBOX_IMAP:
      type_menu = 0;
      server_type = XEP_SERVEREDITOR_IMAP;
      break;
    case XEP_MAILBOX_MBOX:
      type_menu = 1;
      break;
    case XEP_MAILBOX_MH:
      type_menu = 2;
      break;
    case XEP_MAILBOX_POP3:
      type_menu = 3;
      server_type = XEP_SERVEREDITOR_POP3;
      break;
    case XEP_MAILBOX_NNTP:
      type_menu = 4;
      server_type = XEP_SERVEREDITOR_NNTP;
      break;
    default:
      type_menu = 0;
      break;      
    }
  gtk_option_menu_set_history(GTK_OPTION_MENU(mbeditor->type_menu), type_menu);

  a_GtkUtils_entry_set_text(mbeditor->name, mailbox->id);

  if (show & XEP_MBEDITOR_SHOW_SERVER)
    {
      a_GtkServerEditor_set_server(mbeditor->seditor, mailbox->server);
      a_GtkServerEditor_set_type(mbeditor->seditor, server_type);
      gtk_widget_show(XEP_SEDITOR_WIDGET(mbeditor->seditor));
    }
  else
    {
      gtk_widget_hide(XEP_SEDITOR_WIDGET(mbeditor->seditor));
    }

  if (show & XEP_MBEDITOR_SHOW_PATH)
    {
      a_GtkUtils_entry_set_text(mbeditor->path_entry, 
				a_Folder_get_path(mailbox->folder));
      gtk_widget_show(mbeditor->path_table);
    }
  else
    {
      gtk_widget_hide(mbeditor->path_table);
    }

  if (show & XEP_MBEDITOR_SHOW_FOLDERS)
    {
      GtkMbEditor_folders_init(mbeditor);
      gtk_widget_show(mbeditor->folders_widget);
      gtk_widget_show(mbeditor->selection_widget);
    }
  else
    {
      gtk_widget_hide(mbeditor->folders_widget);
      gtk_widget_hide(mbeditor->selection_widget);
    }

  if (show & XEP_MBEDITOR_SHOW_SUBSCRIBED)
    {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mbeditor->subscribed_button), 
				   mailbox->subscribed_only);
      gtk_widget_show(mbeditor->subscribed_button);
    }
  else
    {
      gtk_widget_hide(mbeditor->subscribed_button);
    }

  if (show & XEP_MBEDITOR_SHOW_NAMESPACE)
    {
      gtk_widget_show(mbeditor->namespace_widget);
    }
  else
    {
      gtk_widget_hide(mbeditor->namespace_widget);
    }
  
  if (show & XEP_MBEDITOR_SHOW_CONNECTIONS)
    {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mbeditor->connection_button),
				   mailbox->use_multiple_connections);
      gtk_spin_button_set_value(GTK_SPIN_BUTTON(mbeditor->max_connections),
				mailbox->max_connections);
      gtk_widget_show(mbeditor->connection_button);
      gtk_widget_show(mbeditor->connections_widget);
    }
  else
    {
      gtk_widget_hide(mbeditor->connection_button);
      gtk_widget_hide(mbeditor->connections_widget);
    }

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mbeditor->use_cache),
			       (mailbox->cache || mailbox->flags));
  a_GtkUtils_entry_set_text(mbeditor->cache_entry, mailbox->cache);
  a_GtkUtils_entry_set_text(mbeditor->flags_entry, mailbox->flags);
}

XePMailbox *
a_GtkMbEditor_get_mailbox(XePGtkMbEditor *mbeditor)
{
  return mbeditor->mailbox;
}

void
a_GtkMbEditor_set_ok_callback(XePGtkMbEditor *mbeditor,
			      void (*user_cb)(XePGtkMbEditor *ambeditor,
					      void *auser_data),
			      void *user_data)
{
  mbeditor->user_cb = user_cb;
  mbeditor->user_data = user_data;
}

GtkWidget *
GtkMbEditor_type_menu_create(XePGtkMbEditor *mbeditor)
{
  GtkItemFactory *item_factory;
  GtkItemFactoryEntry entries[] = 
    {
      {_("/IMAP"), NULL, GtkMbEditor_set_type_cb, XEP_MAILBOX_IMAP, NULL},
      {_("/mbox"), NULL, GtkMbEditor_set_type_cb, XEP_MAILBOX_MBOX, NULL},
      {_("/MH"),   NULL, GtkMbEditor_set_type_cb, XEP_MAILBOX_MH,   NULL},
      {_("/POP3"), NULL, GtkMbEditor_set_type_cb, XEP_MAILBOX_POP3, NULL},
      {_("/NNTP"), NULL, GtkMbEditor_set_type_cb, XEP_MAILBOX_NNTP, NULL}
    };

  item_factory = gtk_item_factory_new(GTK_TYPE_OPTION_MENU,
				      "<XetPanMailboxType>/",
				      NULL);
  gtk_item_factory_create_items(item_factory, 
				sizeof(entries)/sizeof(entries[0]),
				entries,
				mbeditor);
  return gtk_item_factory_get_widget(item_factory, "<XetPanMailboxType>/");
}

static void
GtkMbEditor_append_location_page(XePGtkMbEditor *mbeditor)
{
  GtkWidget *label;
  GtkWidget *table;
  GtkWidget *vbox;
  GtkWidget *button;
  int i = 0;

  A_LOGGING_TRACE;
  
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_set_border_width(GTK_CONTAINER(vbox), 12);

  /* Name and type */

  table = gtk_table_new(3, 2, FALSE);
  gtk_table_set_row_spacings(GTK_TABLE(table), 6);
  gtk_table_set_col_spacings(GTK_TABLE(table), 12);

  gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, FALSE, 0);
  
  mbeditor->name = a_GtkUtils_table_add_entry(GTK_TABLE(table), _("_Name:"), 
					      NULL, NULL, i, 0);
  
  i++;

  label = gtk_label_new_with_mnemonic(_("T_ype:"));
  gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, i, i + 1, 
		   GTK_FILL, GTK_SHRINK, 0, 0);
  gtk_widget_show(label);

  mbeditor->type_menu = GtkMbEditor_type_menu_create(mbeditor);
  gtk_label_set_mnemonic_widget(GTK_LABEL(label), mbeditor->type_menu);
  gtk_widget_show(mbeditor->type_menu);
  gtk_table_attach(GTK_TABLE(table), mbeditor->type_menu, 1, 2, i, i + 1, 
		   GTK_FILL | GTK_EXPAND, GTK_SHRINK, 0, 0);

  gtk_widget_show(table);

  /* Server */

  i = 0;

  mbeditor->seditor = a_GtkServerEditor_new();
  gtk_box_pack_start(GTK_BOX(vbox), XEP_SEDITOR_WIDGET(mbeditor->seditor),
		     FALSE, FALSE, 0);

  /* MH and MBOX paths */

  i = 0;
  table = gtk_table_new(1, 3, FALSE);
  gtk_table_set_row_spacings(GTK_TABLE(table), 6);
  gtk_table_set_col_spacings(GTK_TABLE(table), 12);

  gtk_box_pack_start(GTK_BOX(vbox), table, FALSE, FALSE, 0);

  mbeditor->path_entry = a_GtkUtils_table_add_entry(GTK_TABLE(table),
						    _("Mailbox _path:"),
						    NULL, NULL, 1, i);
  button = gtk_button_new_from_stock(GTK_STOCK_OPEN);
  gtk_table_attach(GTK_TABLE(table), button, 3, 4, i, i + 1,
		   GTK_SHRINK, GTK_SHRINK, 0, 0);
  gtk_widget_show(button);
  
  mbeditor->path_table = table;
  
  label = gtk_label_new_with_mnemonic(_("_Location"));
  gtk_widget_show(label);

  gtk_widget_show(vbox);

  gtk_notebook_append_page(GTK_NOTEBOOK(mbeditor->notebook), vbox, label);
}

static void
GtkMbEditor_append_folders_page(XePGtkMbEditor *mbeditor)
{
  GtkWidget *label;
  GtkWidget *hbox;
  GtkWidget *search_box;
  GtkWidget *vbox;
  GtkWidget *vbox1;
  GtkWidget *treeview;
  GtkWidget *button;
  GtkWidget *select_all_button;
  GtkWidget *invert_button;
  GtkWidget *subscribe_button;
  GtkWidget *unsubscribe_button;
  GtkWidget *table;
  GtkWidget *icon;
  int i = 0;

  hbox = gtk_hbox_new(FALSE, 12);
  gtk_container_set_border_width(GTK_CONTAINER(hbox), 12);

  vbox = gtk_vbox_new(FALSE, 6);
  gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, TRUE, 0);
  gtk_widget_show(vbox);

  mbeditor->xtreeview = treeview = gtk_xtree_view_new();
  gtk_xtree_view_insert_column(GTK_XTREE_VIEW(treeview), 0,
			       XEP_MBEDITOR_FOLDER_COL,
			       GTK_XCOLUMN_TYPE_TEXT,
			       TRUE,
			       GTK_SORT_ASCENDING,
			       NULL,
			       _("Folders"));
  gtk_xtree_view_set_expansible_column(GTK_XTREE_VIEW(treeview), 
				       XEP_MBEDITOR_FOLDER_COL);

  icon = gtk_image_new_from_stock(GTK_STOCK_YES, GTK_ICON_SIZE_MENU);
  gtk_xtree_view_insert_column(GTK_XTREE_VIEW(treeview), 0, 
			       XEP_MBEDITOR_SUBSCRIBED_COL,
			       GTK_XCOLUMN_TYPE_STOCKID,
			       TRUE,
			       GTK_SORT_ASCENDING,
			       icon,
			       NULL);
  gtk_xtree_view_set_column_width(GTK_XTREE_VIEW(treeview), 
				  XEP_MBEDITOR_SUBSCRIBED_COL,
				  3*8);

  gtk_xtree_view_set_row_data_func(GTK_XTREE_VIEW(treeview),
				   GtkMbEditor_folder_data_func,
				   mbeditor);
  gtk_xtree_view_set_row_params_func(GTK_XTREE_VIEW(treeview),
				     GtkMbEditor_folder_params_func,
				     mbeditor);
  gtk_xtree_view_set_data_iterator_func(GTK_XTREE_VIEW(treeview),
					GtkMbEditor_folder_iterator_func,
					mbeditor);
  gtk_xtree_view_set_match_func(GTK_XTREE_VIEW(treeview),
				GtkMbEditor_folder_match_func,
				mbeditor);

  gtk_box_pack_start(GTK_BOX(vbox), treeview, TRUE, TRUE, 0);
  gtk_widget_show(treeview);

  mbeditor->folder_button = 
    gtk_button_new_with_mnemonic(_("Download Folder _List"));
  gtk_widget_show(mbeditor->folder_button);
  gtk_box_pack_start(GTK_BOX(vbox), mbeditor->folder_button, FALSE, FALSE, 0);

  search_box = gtk_hbox_new(FALSE, 12);
  gtk_box_pack_start(GTK_BOX(vbox), search_box, FALSE, FALSE, 0);

  label = gtk_label_new_with_mnemonic(_("Searc_h:"));
  gtk_box_pack_start(GTK_BOX(search_box), label, FALSE, FALSE, 0);
  gtk_widget_show(label);

  mbeditor->search_entry = gtk_entry_new();
  gtk_label_set_mnemonic_widget(GTK_LABEL(label), mbeditor->search_entry);
  gtk_box_pack_start(GTK_BOX(search_box), mbeditor->search_entry,
		     FALSE, FALSE, 0);
  gtk_widget_show(mbeditor->search_entry);

  mbeditor->search_button = gtk_button_new_from_stock(GTK_STOCK_FIND);
  gtk_box_pack_start(GTK_BOX(search_box), mbeditor->search_button,
		     FALSE, FALSE, 0);
  gtk_widget_set_sensitive(mbeditor->search_button, FALSE);
  gtk_widget_show(mbeditor->search_button);

  gtk_widget_show(search_box);
  
  mbeditor->folders_widget = vbox;

  /* Control vbox */

  vbox1 = gtk_vbox_new(FALSE, 12);
  gtk_box_pack_start(GTK_BOX(hbox), vbox1, FALSE, FALSE, 0);
  gtk_widget_show(vbox1);

  /* Namespaces */

  vbox = gtk_vbox_new(FALSE, 6);
  gtk_box_pack_start(GTK_BOX(vbox1), vbox, FALSE, FALSE, 0);
  gtk_widget_show(vbox);
  mbeditor->namespace_widget = vbox;

  mbeditor->personal_ns_button = 
    gtk_check_button_new_with_mnemonic(_("Show _Personal Namespace"));
  gtk_box_pack_start(GTK_BOX(vbox), mbeditor->personal_ns_button, 
		     FALSE, FALSE, 0);
  gtk_widget_show(mbeditor->personal_ns_button);

  mbeditor->users_ns_button = 
    gtk_check_button_new_with_mnemonic(_("Show Use_rs Namespace"));
  gtk_box_pack_start(GTK_BOX(vbox), mbeditor->users_ns_button, 
		     FALSE, FALSE, 0);
  gtk_widget_show(mbeditor->users_ns_button);

  mbeditor->shared_ns_button = 
    gtk_check_button_new_with_mnemonic(_("Show Share_d Namespace"));
  gtk_box_pack_start(GTK_BOX(vbox), mbeditor->shared_ns_button, 
		     FALSE, FALSE, 0);
  gtk_widget_show(mbeditor->shared_ns_button);

  button = gtk_button_new_with_mnemonic(_("_Get Namespaces from Server"));
  gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 0);
  gtk_widget_show(button);

  /* Selection and subscription table */

  table = gtk_table_new(4, 2, FALSE);
  gtk_table_set_row_spacings(GTK_TABLE(table), 6);
  gtk_table_set_col_spacings(GTK_TABLE(table), 12);
  gtk_box_pack_end(GTK_BOX(vbox1), table, FALSE, FALSE, 0);

  /* Show only subscribed or selected */

  mbeditor->subscribed_button = 
    gtk_check_button_new_with_mnemonic(_("Show Only Su_bscribed"));
  gtk_table_attach(GTK_TABLE(table), mbeditor->subscribed_button, 0, 2, 
		   i, i + 1, GTK_FILL | GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show(mbeditor->subscribed_button);

  i++;
  mbeditor->selected_button = 
    gtk_check_button_new_with_mnemonic(_("Show Only Se_lected"));
  gtk_table_attach(GTK_TABLE(table), mbeditor->selected_button, 0, 2, 
		   i, i + 1, GTK_FILL | GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show(mbeditor->selected_button);

  mbeditor->selection_widget = table;

  /* Select/Unselect all */

  i++;
  select_all_button = gtk_button_new_with_mnemonic(_("Select _All"));
  gtk_table_attach(GTK_TABLE(table), select_all_button, 0, 1, 
		   i, i + 1, GTK_FILL | GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show(select_all_button);

  invert_button = gtk_button_new_with_mnemonic(_("_Invert Selection"));
  gtk_table_attach(GTK_TABLE(table), invert_button, 1, 2, 
		   i, i + 1, GTK_FILL | GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show(invert_button);

  i++;
  subscribe_button = gtk_button_new_with_mnemonic(_("_Subscribe"));
  gtk_table_attach(GTK_TABLE(table), subscribe_button, 0, 1, 
		   i, i + 1, GTK_FILL | GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show(subscribe_button);

  unsubscribe_button = gtk_button_new_with_mnemonic(_("_Unsubscribe"));
  gtk_table_attach(GTK_TABLE(table), unsubscribe_button, 1, 2, 
		   i, i + 1, GTK_FILL | GTK_EXPAND, GTK_SHRINK, 0, 0);
  gtk_widget_show(unsubscribe_button);

  gtk_widget_show(table);

  /* Signals */

  g_signal_connect(G_OBJECT(mbeditor->folder_button), "clicked",
		   G_CALLBACK(GtkMbEditor_get_folder_list_cb), mbeditor);

  g_signal_connect(G_OBJECT(select_all_button), "clicked",
		   G_CALLBACK(GtkMbEditor_folders_select_all_cb), mbeditor);
  
  g_signal_connect(G_OBJECT(invert_button), "clicked",
		   G_CALLBACK(GtkMbEditor_folders_invert_cb), mbeditor);
  
  g_signal_connect(G_OBJECT(subscribe_button), "clicked",
		   G_CALLBACK(GtkMbEditor_folders_subscribe_cb), mbeditor);

  g_signal_connect(G_OBJECT(unsubscribe_button), "clicked",
		   G_CALLBACK(GtkMbEditor_folders_unsubscribe_cb), mbeditor);

  g_signal_connect_after(G_OBJECT(mbeditor->subscribed_button), "toggled",
			 G_CALLBACK(GtkMbEditor_folders_subscribed_only_cb), 
			 mbeditor);

  g_signal_connect(G_OBJECT(mbeditor->search_entry), "changed", 
		   G_CALLBACK(GtkMbEditor_folder_search_changed_cb),
		   mbeditor);

  g_signal_connect(G_OBJECT(mbeditor->search_entry), "key-press-event", 
		   G_CALLBACK(GtkMbEditor_folder_search_key_press_cb),
		   mbeditor);

  g_signal_connect(G_OBJECT(mbeditor->search_button), "clicked",
		   G_CALLBACK(GtkMbEditor_folder_search_next_cb),
		   mbeditor);

  /* Page label */

  label = gtk_label_new_with_mnemonic(_("_Folders"));
  gtk_widget_show(label);

  gtk_widget_show(hbox);
  gtk_notebook_append_page(GTK_NOTEBOOK(mbeditor->notebook), hbox, label);
}

static void
GtkMbEditor_append_options_page(XePGtkMbEditor *mbeditor)
{
  GtkWidget *label;
  GtkWidget *vbox;

  A_LOGGING_TRACE;

  vbox = gtk_vbox_new(FALSE, 6);
  gtk_widget_show(vbox);

  label = gtk_label_new_with_mnemonic(_("Opt_ions"));
  gtk_widget_show(label);

  gtk_notebook_append_page(GTK_NOTEBOOK(mbeditor->notebook), vbox, label);
}

static void
GtkMbEditor_append_performance_page(XePGtkMbEditor *mbeditor)
{
  GtkWidget *table;
  GtkWidget *label;
  GtkWidget *button;
  GtkWidget *hbox;

  int i;

  A_LOGGING_TRACE;

  table = gtk_table_new(4, 4, FALSE);
  gtk_container_set_border_width(GTK_CONTAINER(table), 12);
  gtk_table_set_row_spacings(GTK_TABLE(table), 6);
  gtk_table_set_col_spacings(GTK_TABLE(table), 12);

  i = 0;

  mbeditor->use_cache =
    gtk_check_button_new_with_mnemonic(_("Use _local cache for faster access"));
  gtk_widget_show(mbeditor->use_cache);

  gtk_table_attach(GTK_TABLE(table), mbeditor->use_cache, 0, 4, i, i + 1,
		   GTK_FILL, GTK_SHRINK, 0, 0);

  i++;
  mbeditor->cache_entry = a_GtkUtils_table_add_entry(GTK_TABLE(table),
						     _("_Cache directory:"),
						     NULL, NULL, 1, i);
  button = gtk_button_new_from_stock(GTK_STOCK_OPEN);
  gtk_table_attach(GTK_TABLE(table), button, 3, 4, i, i + 1,
		   GTK_SHRINK, GTK_SHRINK, 0, 0);
  gtk_widget_show(button);

  i++;
  mbeditor->flags_entry = a_GtkUtils_table_add_entry(GTK_TABLE(table),
						     _("_Flags directory:"), 
						     NULL, NULL, 1, i);
  button = gtk_button_new_from_stock(GTK_STOCK_OPEN);
  gtk_table_attach(GTK_TABLE(table), button, 3, 4, i, i + 1,
		   GTK_SHRINK, GTK_SHRINK, 0, 0);
  gtk_widget_show(button);

  i++;
  mbeditor->connection_button =
    gtk_check_button_new_with_mnemonic(_("_Use one connection for each opened "
					 "folder"));
  gtk_table_attach(GTK_TABLE(table), mbeditor->connection_button,
		   0, 4, i, i + 1,
		   GTK_FILL, GTK_SHRINK, 0, 0);
  gtk_widget_show(mbeditor->connection_button);

  i++;
  hbox = gtk_hbox_new(FALSE, 12);
  gtk_table_attach(GTK_TABLE(table), hbox,
		   1, 4, i, i + 1,
		   GTK_FILL | GTK_EXPAND, GTK_SHRINK, 0, 0);
  mbeditor->connections_widget = hbox;
  gtk_widget_show(hbox);

  label = gtk_label_new_with_mnemonic(_("_Maximal number of opened "
					"connections:"));
  gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
  gtk_widget_show(label);

  mbeditor->max_connections = gtk_spin_button_new_with_range(1, 10, 1);
  gtk_spin_button_set_digits(GTK_SPIN_BUTTON(mbeditor->max_connections), 0);
  gtk_label_set_mnemonic_widget(GTK_LABEL(label), mbeditor->max_connections);
  gtk_box_pack_start(GTK_BOX(hbox), mbeditor->max_connections,
		     FALSE, FALSE, 0);
  gtk_widget_show(mbeditor->max_connections);
			     

  label = gtk_label_new_with_mnemonic(_("Per_formance"));
  gtk_widget_show(label);

  gtk_widget_show(table);

  gtk_notebook_append_page(GTK_NOTEBOOK(mbeditor->notebook), table, label);
}

static unsigned int
GtkMbEditor_get_widgets_from_type(XePMailboxType type)
{
  
  switch (type)
    {
    case XEP_MAILBOX_IMAP:
      return XEP_MBEDITOR_SHOW_IMAP;
      break;
    case XEP_MAILBOX_MBOX:
      return XEP_MBEDITOR_SHOW_MBOX;
      break;
    case XEP_MAILBOX_MH:
      return XEP_MBEDITOR_SHOW_MH;
      break;
    case XEP_MAILBOX_NNTP:
      return XEP_MBEDITOR_SHOW_NNTP;
      break;
    case XEP_MAILBOX_POP3:
      return XEP_MBEDITOR_SHOW_POP3;
      break;
    default:
      return 0; /* If unknown type, i.e. if new mailbox, show nothing */
    }
}

static void
GtkMbEditor_set_type_cb(gpointer ambeditor, guint mboxtype, GtkWidget *widget)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;

  GtkMbEditor_apply_changes(mbeditor);
  a_Mailbox_set_type(mbeditor->mailbox, mboxtype);

  GtkMbEditor_show_mailbox(mbeditor);
}

static void
GtkMbEditor_apply_changes(XePGtkMbEditor *mbeditor)
{
  unsigned int show;
  XePMailbox *mailbox = mbeditor->mailbox;

  show = GtkMbEditor_get_widgets_from_type(mailbox->type);

  /* TODO: check if mailbox id is unique */
  x_free(mailbox->id);
  mailbox->id = gtk_editable_get_chars(GTK_EDITABLE(mbeditor->name), 0, -1);

  if (show & XEP_MBEDITOR_SHOW_SERVER)
    {
      a_GtkServerEditor_apply_changes(mbeditor->seditor);
    }

  if (show & XEP_MBEDITOR_SHOW_PATH)
    {
      a_Mailbox_set_toplevel(mailbox, gtk_entry_get_text(GTK_ENTRY(mbeditor->path_entry)));
    }
  
  if (show & XEP_MBEDITOR_SHOW_FOLDERS)
    {
      GtkMbEditor_set_folder_list(mbeditor);
    }

  if (show & XEP_MBEDITOR_SHOW_SUBSCRIBED)
    {
      mailbox->subscribed_only = 
	gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mbeditor->subscribed_button));

    }

  if (show & XEP_MBEDITOR_SHOW_NAMESPACE)
    {
      /* TODO */
    }
  
  if (show & XEP_MBEDITOR_SHOW_CONNECTIONS)
    {
      mailbox->use_multiple_connections = 
	gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mbeditor->connection_button));

      mailbox->max_connections = 
	gtk_spin_button_get_value(GTK_SPIN_BUTTON(mbeditor->max_connections));
    }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(mbeditor->use_cache)))
    {
      const char *cache;
      const char *flags;

      cache = gtk_entry_get_text(GTK_ENTRY(mbeditor->cache_entry));
      flags = gtk_entry_get_text(GTK_ENTRY(mbeditor->flags_entry));

      if (cache[0] == 0x00)
	cache = NULL;
      if (flags[0] == 0x00)
	flags = NULL;

      /* TODO: changing cache options should update opened folders */
      if ((mailbox->cache == NULL && cache) || 
	  (mailbox->cache && cache == NULL) ||
	  (mailbox->cache && cache && strcmp(mailbox->cache, cache)))
	{
	  x_free(mailbox->cache);
	  mailbox->cache = X_STRDUP(cache);
	}
      
      /* TODO: changing cache options should update opened folders */
      if ((mailbox->flags == NULL && flags) || 
	  (mailbox->flags && flags == NULL) ||
	  (mailbox->flags && flags && strcmp(mailbox->flags, flags)))
	{
	  x_free(mailbox->flags);
	  mailbox->flags = X_STRDUP(flags);
	}
    }
  else
    {
      x_free(mailbox->cache);
      x_free(mailbox->flags);
      mailbox->cache = NULL;
      mailbox->flags = NULL;
    }
}

static void
GtkMbEditor_folders_init(XePGtkMbEditor *mbeditor)
{
  gtk_xtree_view_set_length(GTK_XTREE_VIEW(mbeditor->xtreeview),
			    clist_count(mbeditor->folder_list));

  if (mbeditor->folder_list && clist_count(mbeditor->folder_list))
    {
      gtk_xtree_view_set_first_row(GTK_XTREE_VIEW(mbeditor->xtreeview),
				   clist_begin(mbeditor->folder_list));
    }

  gtk_xtree_view_update(GTK_XTREE_VIEW(mbeditor->xtreeview));
}

static gboolean
GtkMbEditor_folder_data_func(void *iterator, 
			     guint col_id, 
			     void **content,
			     gchar **attribute,
			     gpointer data)
{
  clistiter *iter = (clistiter *) iterator;
  XePGtkMbEditorFolder *mbfolder;

  if (iter)
    {
      mbfolder = (XePGtkMbEditorFolder *) iter->data;
      *content = (col_id == XEP_MBEDITOR_FOLDER_COL ? 
		  mbfolder->path :
		  mbfolder->flags & XEP_MBEDITOR_SUBSCRIBED ?
		  GTK_STOCK_YES :
		  NULL);
      *attribute = (mbfolder->flags & XEP_MBEDITOR_SUBSCRIBED ? 
		    "subscribed" :
		    "unsubscribed");
    }
  return TRUE;
}

static void
GtkMbEditor_folder_params_func(void *iterator, gboolean *has_children,
			       guint *depth, gpointer data)
{
  *has_children = FALSE;
  *depth = 0;
}
			       
static gboolean
GtkMbEditor_folder_iterator_func(void **iterator, 
				 gboolean next,
				 gboolean ignore_children,
				 guint sort_col_id,
				 GtkSortType sort_order,
				 gpointer ambeditor)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;
  clistiter *iter = (clistiter *) *iterator;
  XePGtkMbEditorFolder *mbfolder;

  if (iter)
    {
      /* TODO: implement filtering out unsubscribed items */
      iter = (next ? iter->next : iter->previous);
      if (iter)
	{
	  *iterator = iter;
	  return TRUE;
	}
    }
  return FALSE;
}

static gboolean
GtkMbEditor_folder_match_func(void *iterator, gpointer ambeditor)
{
 XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;
 clistiter *iter = (clistiter *) iterator;

 if (iter)
   {
     XePGtkMbEditorFolder *mbfolder = (XePGtkMbEditorFolder *) iter->data;
     if (mbeditor->valid_preg)
       return (regexec(&mbeditor->preg, mbfolder->path, 0, NULL, 0) == 0);
     else
       return TRUE;
   }
 return FALSE;
}

static void
GtkMbEditor_folder_search_changed_cb(GtkEditable *editable, gpointer ambeditor)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;
  const char *expr = gtk_entry_get_text(GTK_ENTRY(mbeditor->search_entry));

  if (mbeditor->valid_preg)
    {
      regfree(&mbeditor->preg);
      mbeditor->valid_preg = FALSE;
    }

  if (expr && *expr)
    {
      mbeditor->valid_preg = (0 == 
			      regcomp(&mbeditor->preg, 
				      expr, 
				      REG_ICASE | REG_NOSUB | REG_EXTENDED));
    }
  gtk_widget_set_sensitive(mbeditor->search_button, mbeditor->valid_preg);
}

static gboolean
GtkMbEditor_folder_search_key_press_cb(GtkEditable *editable, 
				       GdkEventKey *event,
				       gpointer ambeditor)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;

  switch (event->keyval)
    {
    case GDK_Escape:
      gtk_widget_grab_focus(mbeditor->xtreeview);
      return TRUE;
      break;
    case GDK_Return:
      if ((event->state & GDK_MODIFIER_MASK) == 0)
	gtk_xtree_view_next_match(GTK_XTREE_VIEW(mbeditor->xtreeview), TRUE);
      else if (event->state & GDK_SHIFT_MASK)
	gtk_xtree_view_prev_match(GTK_XTREE_VIEW(mbeditor->xtreeview), TRUE);
      return TRUE;
    }
  return FALSE;
}

static void
GtkMbEditor_folder_search_next_cb(GtkWidget *widget, gpointer ambeditor)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;
  
  gtk_xtree_view_next_match(GTK_XTREE_VIEW(mbeditor->xtreeview), TRUE);
}

static void
GtkMbEditor_folders_subscribed_only_cb(GtkButton *button, gpointer ambeditor)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;
}

static void
GtkMbEditor_folders_subscribe_cb(GtkButton *button, gpointer ambeditor)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;
  void **selection;
  uint32_t n, i;

  n = gtk_xtree_view_get_selection(GTK_XTREE_VIEW(mbeditor->xtreeview),
				   &selection);

  if (n)
    {
      gtk_xtree_view_freeze(GTK_XTREE_VIEW(mbeditor->xtreeview));
      for (i = 0; i < n; i++)
	{
	  clistiter *iter = (clistiter *) selection[i];
	  XePGtkMbEditorFolder *mbfolder = (XePGtkMbEditorFolder *) iter->data;

	  mbfolder->flags |= XEP_MBEDITOR_SUBSCRIBED;
	}	    
      free(selection);
      gtk_xtree_view_thaw(GTK_XTREE_VIEW(mbeditor->xtreeview));
    }
}

static void
GtkMbEditor_folders_unsubscribe_cb(GtkButton *button, gpointer ambeditor)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;
  void **selection;
  uint32_t n, i;

  n = gtk_xtree_view_get_selection(GTK_XTREE_VIEW(mbeditor->xtreeview),
				   &selection);

  if (n)
    {
      gtk_xtree_view_freeze(GTK_XTREE_VIEW(mbeditor->xtreeview));
      for (i = 0; i < n; i++)
	{
	  clistiter *iter = (clistiter *) selection[i];
	  XePGtkMbEditorFolder *mbfolder = (XePGtkMbEditorFolder *) iter->data;

	  mbfolder->flags &= ~XEP_MBEDITOR_SUBSCRIBED;
	}	    
      free(selection);
      gtk_xtree_view_thaw(GTK_XTREE_VIEW(mbeditor->xtreeview));
    }
}

static void
GtkMbEditor_folders_select_all_cb(GtkButton *button, gpointer ambeditor)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;

  gtk_xtree_view_select_all(GTK_XTREE_VIEW(mbeditor->xtreeview), NULL);
}

static void
GtkMbEditor_folders_invert_cb(GtkButton *button, gpointer ambeditor)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;
  gtk_xtree_view_invert_selection(GTK_XTREE_VIEW(mbeditor->xtreeview), NULL);
}

static void
GtkMbEditor_get_folder_list_cb(GtkButton *button, gpointer ambeditor)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;
  XePManagerMsg *msg;
  XePMailbox *mailbox = mbeditor->mailbox;
  XePFolder *folder = mailbox->folder;

  mbeditor->block++;

  gtk_widget_set_sensitive(mbeditor->folder_button, FALSE);
  GtkMbEditor_apply_changes(mbeditor);

  if (mailbox->ep_mail_list)
    {
      gtk_xtree_view_reset(GTK_XTREE_VIEW(mbeditor->xtreeview));
      mail_list_free(mailbox->ep_mail_list);
      mailbox->ep_mail_list = NULL;
    }

  GtkMbEditor_folder_list_reset(mbeditor);

  msg = a_Manager_msg_new();
  a_Manager_msg_set_type(msg, XEP_MANAGER_LIST);
  a_Manager_msg_set_source_target(msg, NULL, &folder->thread);
  a_Manager_msg_set_args(msg, folder, &mailbox->ep_mail_list);
  a_Manager_msg_add_callback(msg, GtkMbEditor_folder_list_cb, mbeditor);
  if (mailbox->type != XEP_MAILBOX_IMAP)
    {
      /* If not IMAP, pretend lsub has already finished */
      a_Manager_msg_add_callback(msg, GtkMbEditor_folder_lsub_cb, mbeditor);
    }
  a_Manager_msg_submit(msg);

  if (mailbox->type == XEP_MAILBOX_IMAP)
    {
      if (mailbox->ep_mail_lsub)
	{
	  mail_list_free(mailbox->ep_mail_lsub);
	  mailbox->ep_mail_lsub = NULL;
	}
      msg = a_Manager_msg_new();
      a_Manager_msg_set_type(msg, XEP_MANAGER_LSUB);
      a_Manager_msg_set_source_target(msg, NULL, &folder->thread);
      a_Manager_msg_set_args(msg, folder, &mailbox->ep_mail_lsub);
      a_Manager_msg_add_callback(msg, GtkMbEditor_folder_lsub_cb, mbeditor);
      a_Manager_msg_submit(msg);
    }

  msg = a_Manager_msg_new();
  a_Manager_msg_set_type(msg, XEP_MANAGER_MAILBOX_DISCONNECT);
  a_Manager_msg_set_source_target(msg, NULL, &folder->thread);
  a_Manager_msg_set_args(msg, folder, folder);
  a_Manager_msg_add_callback(msg, GtkMbEditor_folder_finalize_cb, mbeditor);
  a_Manager_msg_submit(msg);

  msg = a_Manager_msg_new();
  a_Manager_msg_set_type(msg, XEP_MANAGER_EXIT);
  a_Manager_msg_set_source_target(msg, NULL, &folder->thread);
  a_Manager_msg_submit(msg);
}

static void
GtkMbEditor_folder_list_reset(XePGtkMbEditor *mbeditor)
{
  gtk_xtree_view_reset(GTK_XTREE_VIEW(mbeditor->xtreeview));

  if (mbeditor->folder_hash)
    {
      chash_free(mbeditor->folder_hash);
      mbeditor->folder_hash = NULL;
    }

  if (mbeditor->folder_list)
    {
      clistiter *iter;
      XePGtkMbEditorFolder *mbfolder;
      for (iter = clist_begin(mbeditor->folder_list); iter; iter = iter->next)
	{
	  mbfolder = (XePGtkMbEditorFolder *) iter->data;
	  free(mbfolder->path);
	  free(mbfolder);
	}
      clist_free(mbeditor->folder_list);
      mbeditor->folder_list = NULL;
    }
}

static void
GtkMbEditor_folder_list_cb(void *ambeditor, void *amb_list)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;

  if (mbeditor->mailbox->ep_mail_list)
    {
      GtkMbEditor_folder_list_register(mbeditor, 
				       mbeditor->mailbox->ep_mail_list->list,
				       0);
    }
}

static void
GtkMbEditor_folder_lsub_cb(void *ambeditor, void *amb_lsub)
{
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;
  
  if (mbeditor->mailbox->ep_mail_lsub)
    {
      GtkMbEditor_folder_list_register(mbeditor, 
				       mbeditor->mailbox->ep_mail_lsub->list,
				       XEP_MBEDITOR_SUBSCRIBED);
    }
}

static void
GtkMbEditor_folder_list_register(XePGtkMbEditor *mbeditor, clist *folder_list,
				 int flag)
{
  int hash_size;
  clistiter *iter;
  clistiter *mbfolder_iter;
  char *folder_path;
  chashdatum key;
  chashdatum value;
  XePGtkMbEditorFolder *mbfolder;
  
  if (mbeditor->folder_hash == NULL)
    {
      hash_size = (XEP_MBEDITOR_CHASH_MAXSIZE > folder_list->count ? 
		   folder_list->count :
		   XEP_MBEDITOR_CHASH_MAXSIZE);
      
      mbeditor->folder_hash = chash_new(hash_size, CHASH_COPYNONE);
    }
  if (mbeditor->folder_list == NULL)
    {
      mbeditor->folder_list = clist_new();
    }

  for (iter = clist_begin(folder_list); iter; iter = iter->next)
    {
      folder_path = (char *) iter->data;
      key.data = folder_path;
      key.len = strlen(folder_path);
      if (chash_get(mbeditor->folder_hash, &key, &value) == 0)
	{
	  mbfolder_iter = (clistiter *) value.data;
	  mbfolder = (XePGtkMbEditorFolder *) mbfolder_iter->data;
	  mbfolder->flags = flag;
	}
      else
	{
	  mbfolder = x_alloc(1, XePGtkMbEditorFolder);
	  mbfolder->path = strdup(folder_path);
	  mbfolder->flags = flag;
	  clist_append(mbeditor->folder_list, mbfolder);

	  value.data = (void *) (clist_end(mbeditor->folder_list));
	  value.len  = sizeof(clistiter *);
	  chash_set(mbeditor->folder_hash, &key, &value, NULL);
	}
    }
}

static void
GtkMbEditor_folder_finalize_cb(void *ambeditor, void *anull)
{	  
  XePGtkMbEditor *mbeditor = (XePGtkMbEditor *) ambeditor;

  GtkMbEditor_folders_init(mbeditor);

  gtk_widget_set_sensitive(mbeditor->folder_button, TRUE);
  mbeditor->block--;
}

static void
GtkMbEditor_set_folder_list(XePGtkMbEditor *mbeditor)
{
  clistiter *iter;
  clistiter *mbfolder_iter;
  XePGtkMbEditorFolder *mbfolder;
  XePMailbox *mailbox = mbeditor->mailbox;
  XePFolder *folder;
  char *folder_path;
  chashdatum key;
  chashdatum value;

  /* First, update the existing folders: 
   * - if an existing folder is subscribed to while it was marked for delete
   *   (negative uid), then rehabilitate
   * - if an existing folder is not subscribed to while it has a valid 
   *   (positive uid), then mark it for delete
   * At the end of this loop, existing folders are marked for deleted
   * or are marked to remain
   */
  for (iter = clist_begin(mailbox->folders); iter; iter = iter->next)
    {
      folder = (XePFolder *) iter->data;
      folder_path = folder->path;
      key.data = folder_path;
      key.len  = strlen(folder_path);
      if (chash_get(mbeditor->folder_hash, &key, &value) == 0)
	{
	  mbfolder_iter = (clistiter *) value.data;
	  mbfolder = (XePGtkMbEditorFolder *) mbfolder_iter->data;
	  if (((mbfolder->flags & XEP_MBEDITOR_SUBSCRIBED) && 
	       folder->uid < 0) ||
	      (!(mbfolder->flags & XEP_MBEDITOR_SUBSCRIBED) &&
	       folder->uid > 0))
	    
	    {
	      folder->uid = -folder->uid;
	    }
	  mbfolder->flags &= ~XEP_MBEDITOR_SUBSCRIBED;
	}
      else
	{
	  A_LOGGING_SYSTEM_ERROR("unregistered folder path %s", folder_path);
	}
    }

  /* Add new subscribed folders 
   */
  for (iter = clist_begin(mbeditor->folder_list); iter; iter = iter->next)
    {
      mbfolder = (XePGtkMbEditorFolder *) iter->data;
      if (mbfolder->flags & XEP_MBEDITOR_SUBSCRIBED)
	{
	  /* New folders are created and registered */
	  folder = a_Folder_new(XEP_FOLDER_TYPE_REAL);
	  a_Folder_set_path(folder, mbfolder->path);
	  a_Folder_set_mailbox(folder, mailbox);
	  a_Folder_update_id(folder);
	  a_Folder_register(folder);
	  clist_append(mailbox->folders, folder);
	  mbfolder->flags &= ~XEP_MBEDITOR_SUBSCRIBED; /* Not really needed */
	}
    }
}

#endif
