/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef USE_GTKUI

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <gtk/gtkwidget.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtktoolbar.h>
#include <gtk/gtkimage.h>
#include <gtk/gtkstock.h>

#include "intl.h"
#include "logging.h"
#include "x_alloc.h"

#include "ui.h"
#include "gtk-ui.h"
#include "gtk-toolbar.h"

#include "gtk-utils.h"
#include "gtk-log.h"

typedef struct XePGTKToolbarCommand_s XePGTKToolbarCommand;

struct XePGTKToolbarCommand_s
{
  char *name;
  char *tooltip;
  char *private_tooltip;
  char *icon;
  void (*callback)(GtkWidget *widget, gpointer data);
};

struct XePGTKToolbar_s
{
  GtkWidget *widget;
  GtkWidget *command;
  GtkWidget *status;
  GtkWidget *status_error;
  GtkWidget *status_ok;
};

static XePGTKToolbar *
GtkToolbar_new(void);

static void
GtkToolbar_free(XePGTKToolbar *toolbar);

static void
GtkToolbar_add_items(GtkWidget *toolbar,
		     XePGTKToolbarCommand *items,
		     gpointer data);

static void
GtkToolbar_status_create(XePUIMain *uimain, XePGTKToolbar *uitoolbar);

static void
GtkToolbar_refresh_cb(GtkWidget *widget, gpointer data);

static void
GtkToolbar_go_back_cb(GtkWidget *widget, gpointer data);

static void
GtkToolbar_go_forward_cb(GtkWidget *widget, gpointer data);

static void
GtkToolbar_find_cb(GtkWidget *widget, gpointer data);

static void
GtkToolbar_print_cb(GtkWidget *widget, gpointer data);

static void
GtkToolbar_compose_cb(GtkWidget *widget, gpointer data);

static void
GtkToolbar_reply_cb(GtkWidget *widget, gpointer data);

static void
GtkToolbar_forward_cb(GtkWidget *widget, gpointer data);

static void
GtkToolbar_delete_cb(GtkWidget *widget, gpointer data);

static void
GtkToolbar_expunge_cb(GtkWidget *widget, gpointer data);

static void
GtkToolbar_status_cb(GtkWidget *widget, gpointer data);

static void
GtkToolbar_send_cb(GtkWidget *widget, gpointer data);

static void
GtkToolbar_draft_cb(GtkWidget *widget, gpointer data);

static void
GtkToolbar_add_cc_cb(GtkWidget *widget, gpointer data);

static void
GtkToolbar_add_bcc_cb(GtkWidget *widget, gpointer data);

static XePGTKToolbarCommand uimain_toolbar[] =
  {
    {
      N_("Refresh"), 
      N_("Refresh folder"), 
      N_("Refresh folder to check for changes"),
      GTK_STOCK_REFRESH,
      GtkToolbar_refresh_cb
    },
    {
      N_("Previous"), 
      N_("Previous matching message"), 
      N_("Search backwards for matching message"),
      GTK_STOCK_GO_BACK,
      GtkToolbar_go_back_cb
    },
    {
      N_("Next"),
      N_("Next matching message"),
      N_("Search forwards for matching message"),
      GTK_STOCK_GO_FORWARD,
      GtkToolbar_go_forward_cb
    },
    {
      N_("Find"),
      N_("Find a message"),
      N_("Find a message matching some conditions"),
      GTK_STOCK_FIND,
      GtkToolbar_find_cb
    },
    {
      N_("Print"),
      N_("Print selected messages"),
      N_("Print all the selected messages"),
      GTK_STOCK_PRINT,
      GtkToolbar_print_cb
    },
    {
      N_("Compose"),
      N_("Compose a new message"),
      N_("Compose a new message"),
      GTK_STOCK_NEW,
      GtkToolbar_compose_cb
    },
    {
      N_("Reply"),
      N_("Reply to message"),
      N_("Reply to the selected messages"),
      XEP_STOCK_REPLY,
      GtkToolbar_reply_cb
    },
    {
      N_("Forward"),
      N_("Forward a message"),
      N_("Forward a message to a third person"),
      XEP_STOCK_FORWARD,
      GtkToolbar_forward_cb
    },
    {
      "Space",
      NULL, 
      NULL,
      NULL,
      NULL,
    },
    {
      N_("Delete"),
      N_("Delete a message"),
      N_("Delete the selected messages."),
      GTK_STOCK_DELETE,
      GtkToolbar_delete_cb
    },
    {
      N_("Expunge"),
      N_("Expunge folder"),
      N_("Expunge folder and completly erase deleted messages."),
      GTK_STOCK_EXECUTE,
      GtkToolbar_expunge_cb
    },
    {
      NULL,
      NULL,
      NULL,
      NULL,
      NULL
    }
  };

XePGTKToolbarCommand uicompose_toolbar[] =
  {
    {
      N_("Send"),
      N_("Send the message now"),
      N_("Send the message now"),
      GTK_STOCK_JUMP_TO,
      GtkToolbar_send_cb
    },
    {
      N_("Draft"),
      N_("Save the message as draft"),
      N_("Save the message as draft and close compose window"),
      GTK_STOCK_SAVE,
      GtkToolbar_draft_cb
    },
    {
      N_("Cc"),
      N_("Add Cc"),
      N_("Add a new Cc header"),
      GTK_STOCK_ADD,
      GtkToolbar_add_cc_cb
    },
    {
      N_("Bcc"),
      N_("Add Bcc"),
      N_("Add a new Bcc header"),
      GTK_STOCK_ADD,
      GtkToolbar_add_bcc_cb
    },
    {
      NULL,
      NULL,
      NULL,
      NULL,
      NULL
    }
  };


/* Function definitions */

static XePGTKToolbar *
GtkToolbar_new(void)
{
  XePGTKToolbar *uitoolbar;
  GtkWidget *hbox;
  GtkWidget *toolbar;

  uitoolbar = x_alloc(1, XePGTKToolbar);
  if (uitoolbar == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
      return NULL;
    }

  hbox =  gtk_hbox_new(FALSE, 0);
  uitoolbar->widget = hbox;
  
  /* Command Toolbar */
  toolbar = gtk_toolbar_new();
  uitoolbar->command = toolbar;
  gtk_toolbar_set_style(GTK_TOOLBAR(toolbar), GTK_TOOLBAR_ICONS);
  gtk_box_pack_start(GTK_BOX(hbox), toolbar, TRUE, TRUE, 0);
  gtk_widget_show(toolbar);

  /* Status Toolbar */
  toolbar = gtk_toolbar_new();
  uitoolbar->status = toolbar;
  gtk_toolbar_set_style(GTK_TOOLBAR(toolbar), GTK_TOOLBAR_ICONS);
  gtk_box_pack_end(GTK_BOX(hbox), toolbar, FALSE, FALSE, 0);
  gtk_widget_show(toolbar);

  return uitoolbar;
}

static void
GtkToolbar_free(XePGTKToolbar *toolbar)
{
  x_free(toolbar);
}


static void
GtkToolbar_add_items(GtkWidget *toolbar,
		     XePGTKToolbarCommand *items,
		     gpointer data)
{
  XePGTKToolbarCommand *item;
  GtkWidget *icon;

  for (item = items; item->name; item++)
    {
      if (strcmp(item->name, "Space") == 0)
	{
	  gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));
	}
      else
	{
	  icon = gtk_image_new_from_stock(item->icon, 
					  GTK_ICON_SIZE_SMALL_TOOLBAR);
	  gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
				  gettext(item->name),
				  gettext(item->tooltip),
				  gettext(item->private_tooltip),
				  icon,
				  GTK_SIGNAL_FUNC(item->callback),
				  data);
	}
    }
}

static void
GtkToolbar_status_create(XePUIMain *uimain, XePGTKToolbar *uitoolbar)
{
  GtkWidget *widget;
  GtkWidget *icon;
  GtkWidget *toolbar = uitoolbar->status;

  icon = gtk_image_new_from_stock("xetpan-status", 
				  GTK_ICON_SIZE_SMALL_TOOLBAR);

  widget = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
				   _("Status"),
				   _("Status OK"),
				   _("View Log File to check for "
				     "error messages"),
				   icon,
				   GTK_SIGNAL_FUNC(GtkToolbar_status_cb),
				   uimain);
  
  uitoolbar->status_ok = widget;
  gtk_widget_show(widget);

  icon = gtk_image_new_from_stock("gtk-dialog-error", 
				  GTK_ICON_SIZE_SMALL_TOOLBAR);
  widget = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar),
				   _("Status"),
				   _("Error: Click to view log"),
				   _("View Log File to check for "
				     "error messages"),
				   icon,
				   GTK_SIGNAL_FUNC(GtkToolbar_status_cb),
				   uimain);
  uitoolbar->status_error = widget;
  gtk_widget_hide(widget);
}


static void
GtkToolbar_refresh_cb(GtkWidget *widget, gpointer data)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XePGTKSummaryView *sview;

  sview = (XePGTKSummaryView *) uimain->summaryview;
  if (sview->folder_ref->ref)
    {  
      a_UI_update_folder(sview->folder_ref->ref);
    }
}

static void
GtkToolbar_go_back_cb(GtkWidget *widget, gpointer data)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;

  A_LOGGING_TRACE;
  a_GtkUI_summaryview_find_match(sview, FALSE);
}

static void
GtkToolbar_go_forward_cb(GtkWidget *widget, gpointer data)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XePGTKSummaryView *sview = (XePGTKSummaryView *) uimain->summaryview;

  A_LOGGING_TRACE;
  a_GtkUI_summaryview_find_match(sview, TRUE);
}

static void
GtkToolbar_find_cb(GtkWidget *widget, gpointer data)
{
  A_LOGGING_TRACE;
  A_LOGGING_SYSTEM_ERROR("Not implemented yet");
}

static void
GtkToolbar_print_cb(GtkWidget *widget, gpointer data)
{
  A_LOGGING_TRACE;
  A_LOGGING_SYSTEM_ERROR("Not implemented yet");
}

static void
GtkToolbar_compose_cb(GtkWidget *widget, gpointer data)
{
  XePUIMain *uimain = (XePUIMain *) data;
  XePUIComposeWindow *uicompose;
  gchar *display;

  A_LOGGING_TRACE;

  display = a_GtkUtils_make_display_name(uimain->toplevel);

  uicompose = a_UI_composewindow_compose(uimain->xetpan, display, 
					 XETPAN_UI_GTK, NULL);

  g_free(display);
}

static void
GtkToolbar_reply_cb(GtkWidget *widget, gpointer data)
{
  XePUIMain *uimain = (XePUIMain *) data;

  A_LOGGING_TRACE;

  a_GtkUI_mainview_reply(uimain);
}

static void
GtkToolbar_forward_cb(GtkWidget *widget, gpointer data)
{
  XePUIMain *uimain = (XePUIMain *) data;

  A_LOGGING_TRACE;

  a_GtkUI_mainview_forward(uimain);
}

static void
GtkToolbar_delete_cb(GtkWidget *widget, gpointer data)
{
  /* #if ENABLE_DELETE_ONLY_WHEN_UI_FEEDBACK_IS_COMPLETE */
  XePUIMain *uimain = (XePUIMain *) data;
  A_LOGGING_TRACE;
  a_GtkUI_summaryview_selection_delete(uimain, XEP_FLAGS_TOGGLE);
  /* #endif */
}

static void
GtkToolbar_expunge_cb(GtkWidget *widget, gpointer data)
{
  XePUIMain *uimain = (XePUIMain *) data;
  A_LOGGING_TRACE;
  a_GtkUI_summaryview_expunge(uimain);
}

static void
GtkToolbar_status_cb(GtkWidget *widget, gpointer data)
{
  XePUIMain *uimain = (XePUIMain *) data;

  A_LOGGING_TRACE;
  
  a_GtkLog_show((XePGTKLog *) uimain->logview);
}

static void
GtkToolbar_send_cb(GtkWidget *widget, gpointer data)
{
  XePGTKComposeWindow *uicompose = (XePGTKComposeWindow *) data;

  A_LOGGING_TRACE;

  a_GtkComposewindow_send(uicompose);
}

static void
GtkToolbar_draft_cb(GtkWidget *widget, gpointer data)
{
  A_LOGGING_TRACE;
  A_LOGGING_SYSTEM_ERROR("Not implemented yet");
}

static void
GtkToolbar_add_cc_cb(GtkWidget *widget, gpointer data)
{
  A_LOGGING_TRACE;
  a_GtkComposewindow_add_cc_header(data, NULL);
}

static void
GtkToolbar_add_bcc_cb(GtkWidget *widget, gpointer data)
{
  A_LOGGING_TRACE;
  a_GtkComposewindow_add_bcc_header(data, NULL);
}

GtkWidget *
a_GtkToolbar_mainview_new(XePUIMain *uimain)
{
  XePGTKToolbar *uitoolbar;
  
  uitoolbar = GtkToolbar_new();

  GtkToolbar_add_items(uitoolbar->command, uimain_toolbar, uimain);
  
  GtkToolbar_status_create(uimain, uitoolbar);

  uimain->toolbar = uitoolbar;

  return uitoolbar->widget;
}

void
a_GtkToolbar_mainview_free(XePUIMain *uimain)
{
  GtkToolbar_free((XePGTKToolbar *) uimain->toolbar);
}

GtkWidget *
a_GtkToolbar_composewindow_new(XePGTKComposeWindow *uicompose)
{
  XePGTKToolbar *uitoolbar;
  
  uitoolbar = GtkToolbar_new();

  GtkToolbar_add_items(uitoolbar->command, uicompose_toolbar, uicompose);

  uicompose->toolbar = uitoolbar;

  return uitoolbar->widget;
}

void
a_GtkToolbar_composewindow_free(XePGTKComposeWindow *uicompose)
{
  GtkToolbar_free(uicompose->toolbar);
}

void
a_GtkToolbar_status_notify_error(XePUIMain *uimain)
{
  XePGTKToolbar *uitoolbar = (XePGTKToolbar *) uimain->toolbar;
  gtk_widget_hide(uitoolbar->status_ok);
  gtk_widget_show(uitoolbar->status_error);
}

void
a_GtkToolbar_status_notify_error_reset(XePUIMain *uimain)
{
  XePGTKToolbar *uitoolbar = (XePGTKToolbar *) uimain->toolbar;
  gtk_widget_hide(uitoolbar->status_error);
  gtk_widget_show(uitoolbar->status_ok);
}

#endif
