/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef IDENTITY_H
#define IDENTITY_H

#include <clist.h>

typedef struct XePIdentity_s      XePIdentity;

typedef void (*XePIdentityDisconnectAllCb)(XePIdentity *identity, void *data);

/*! Identity data structure.
 *  This contains all necessary data for an identity.
 *  An identity is a user name with an address email to which correspond
 *  one or more mailboxes (mail sources).
 */
struct XePIdentity_s
{
  
  char  *id;         /**< Unique identifier. FIXME: Should be an integer? */
  char  *name;       /**< User name (e.g. "John Smith").                  */
  char  *email;      /**< The email address of the user.                  */
  char  *organization; /**< The organization                              */
  char  *reply_to;   /**< Reply-To                                        */
  clist *signatures; /**< List of signatures.                             */
  clist *mailboxes  ;/**< List of XePMailbox mailboxes.                   */

  char *draft_ref;
  char *queue_ref;
  char *sent_ref;
  char* mta_ref;
  char *templates_ref;
};

clist *
a_Identity_list_new(void);

void
a_Identity_list_free(clist *identities);

clist *
a_Identity_list_get(void);

XePIdentity *
a_Identity_new(void);

XePIdentity *
a_Identity_new_copy(XePIdentity *identity);

void
a_Identity_free(XePIdentity *identity);

void
a_Identity_free_copy(XePIdentity *identity);

/*! Set the ID string used for this identity */
void
a_Identity_set_id(XePIdentity *identity, const char* id);

const char *
a_Identity_get_id(XePIdentity *identity);

void
a_Identity_set_name(XePIdentity *identity, const char* name);

const char *
a_Identity_get_name(XePIdentity *identity);

void
a_Identity_set_email(XePIdentity *identity, const char* email);

const char *
a_Identity_get_email(XePIdentity *identity);

void
a_Identity_set_organization(XePIdentity *identity, const char* organization);

const char *
a_Identity_get_organization(XePIdentity *identity);

void
a_Identity_set_reply_to(XePIdentity *identity, const char* reply_to);

const char *
a_Identity_get_reply_to(XePIdentity *identity);

void
a_Identity_set_draft_ref(XePIdentity *identity, const char* draft_ref);

const char *
a_Identity_get_draft_ref(XePIdentity *identity);

void
a_Identity_set_queue_ref(XePIdentity *identity, const char* queue_ref);

const char *
a_Identity_get_queue_ref(XePIdentity *identity);

void
a_Identity_set_sent_ref(XePIdentity *identity, const char* sent_ref);

const char *
a_Identity_get_sent_ref(XePIdentity *identity);

void
a_Identity_set_mta_ref(XePIdentity *identity, const char* mta_ref);

const char *
a_Identity_get_mta_ref(XePIdentity *identity);

void
a_Identity_set_templates_ref(XePIdentity *identity, const char* templates_ref);

const char *
a_Identity_get_templates_ref(XePIdentity *identity);

clist *
a_Identity_signatures_new_copy(clist *signatures);

void
a_Identity_set_signatures(XePIdentity *identity, clist *signatures);

clist *
a_Identity_get_signatures(XePIdentity *identity);

void
a_Identity_set_mailboxes(XePIdentity *identity, clist *mailboxes);

clist *
a_Identity_get_mailboxes(XePIdentity *identity);

/*! Find an identity corresponding to an email address
 *  \param identities clist of registered XePIdentit(y)ies
 *  \param email a string containing an email address that will
 *         be used to find the corresponding identity if any
 *  \return the identity corresponding the the email address, or 
 *          the first registered identity if email is NULL.
 *          Otherwise, it returns NULL if identities is NULL or if no
 *           identity was found.
 */
XePIdentity *
a_Identity_lookup_by_email(clist *identities, const char *email);

/*! Check if one of the folders pertaining to this identity is 
 *  currently used
 *  \param identity the identity checked
 *  \return -1 if at least one of the folders of the identity is currently
 *          busy, 0 if none of them are used
 */
int
a_Identity_is_busy(XePIdentity *identity);

void
a_Identity_disconnect_all(XePIdentity *identity, 
			  XePIdentityDisconnectAllCb user_func,
			  void *user_data);

#endif
