/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef TTYUI_H
#define TTYUI_H

#include <ncurses.h>
#include <panel.h>

#include "xetpan.h"
#include "ui-types.h"
#include "mta.h"

typedef struct XePTTYFoldersView_s XePTTYFoldersView;

struct XePTTYFoldersView_s
{
  WINDOW *fvl_pad;      /**< FoldersViewsList pad    */
  WINDOW *fvl_window;   /**< FoldersViewsList window */
  PANEL  *fvl_panel;    /**< FoldersViewsList panel  */

  WINDOW *fv_pad;       /**< Folders of the View pad    */
  WINDOW *fv_window;    /**< Folders of the View window */
  PANEL *fv_panel;      /**< Folders of the View panel  */
};

typedef struct XePTTYSummaryView_s XePTTYSummaryView;

struct XePTTYSummaryView_s
{
};

typedef struct XePTTYMessageView_s XePTTYMessageView;

struct XePTTYMessageView_s
{
  /*! The following members must be the same as the ones in
   *  XePUIMessageViewAny defined in ui.c 
   */
  XePUIType type;      /**< UI type */
  XePUID uid;          /**< Unique Identifier for the message view */
  int editable;        /**< Wether editing is allowed */
  XetPan *xetpan;      /**< The global XetPan data structure */
  XePUIMain *uimain;   /**< The parent main window, if any */
  XePFolder *folder;   /**< The parent folder, if any */
  XePMessage *message; /**< Currently displayed message */
  uint32_t muid;       /**< muid of the currently displayed message, needed
			* because message are displayed in chunks,
			* asychronously (incremental reflow)
			*/
  clist *inlined_list; /**< List of XePUtilsInlinedData describing the
			*   inlined data
			*/
  char *body_text;  /**< Body text of composed message */
  char *default_charset; /**< Body's default charset */

  /* Private data for the TTY UI */
};

typedef struct XePTTYComposeWindow_s XePTTYComposeWindow;

struct XePTTYComposeWindow_s
{
  XePUIType type;      /**< UI type                */
  XetPan *xetpan;      /**< The global XetPan data structure */
  XePUIMain *uimain;   /**< The parent main window if any */
  XePMTA *mta;         /**< The selected MTA */
  XePTTYMessageView *messageview; /**< The related message view */
  char *sent_folder;   /**< The folder that stores outgoing msg */
  /* The rest of the union is private to TTY */
};


void
a_TtyUI_init(XetPan *xetpan, int *argc, char ***argv);

int
a_TtyUI_init_check(XetPan *xetpan, int *argc, char ***argv);

void
a_TtyUI_main(XetPan *xetpan);

void
a_TtyUI_main_quit(XetPan *xetpan);

int
a_TtyUI_events_pending(XetPan *xetpan);

void
a_TtyUI_main_iteration(XetPan *xetpan);

void
a_TtyUI_first_launch(XetPan *xetpan);

void
a_TtyUI_compose_pending_requests(XetPan *xetpan);

XePUIMain *
a_TtyUI_mainview_create(XetPan *xetpan, char *terminal);

void
a_TtyUI_mainview_close(XePUIMain *uimain);

void
a_TtyUI_foldersview_list_updated(XePUIMain *uimain);

void
a_TtyUI_foldersview_list_refresh(XePUIMain *uimain);

void
a_TtyUI_folder_updated(XePUIMain *uimain, XePFolder *folder);

void 
a_TtyUI_folder_refresh(XePUIMain *uimain, XePFolder *folder);

/*! Function called when a message was modified (flags...) 
 *  When a message is modified (for example its flags),
 *  this function is called for every TtyUI and is responsible
 *  for updating the summary view accordingly
 *  \param uimain
 *  \para message the message that was modified
 */
void
a_TtyUI_message_updated(XePUIMain *uimain, XePMessage *messge);

void
a_TtyUI_messageview_finalize(XePUIMessageView *mview, XePMessage *message);

void
a_TtyUI_messageview_mime_view_show(XePUIMessageView *mview,
				   XePMessage *message);

void
a_TtyUI_messageview_mime_view_block(XePUIMessageView *mview,
				    XePMessage *message);

void
a_TtyUI_messageview_mime_view_unblock(XePUIMessageView *mview,
				      XePMessage *message);

void
a_TtyUI_messageview_mime_show_selected(XePUIMessageView *mview,
				       XePMessage *message);

XePUIComposeWindow *
a_TtyUI_composewindow_create(XetPan *xetpan, char *terminal);

void
a_TtyUI_composewindow_set_message(XePUIComposeWindow *uicompose, 
				  XePMessage *message);

void
a_TtyUI_composewindow_set_body_text(XePUIComposeWindow *uicompose,
				    const char *text,
				    size_t text_size,
				    const char *text_charset);
void
a_TtyUI_composewindow_add_header(XePUIComposeWindow *uicompose,
				 XePMessageHeaderType header_type,
				 const char *header_name,
				 const char *header_value);

void
a_TtyUI_composewindow_close(XePUIComposeWindow *uicompose);

XePUIMessageView*
a_TtyUI_messageview_create_standalone(XetPan *xetpan, 
				      char *terminal);

void
a_TtyUI_messageview_close(XePUIMessageView *mview);

XePTTYMessageView *
a_TtyUI_message_viewer(XetPan *xetpan, 
		       char *terminal, 
		       XePMessage *message, 
		       int async);

void
a_TtyUI_about(XePUIMain *uimain);

void
a_TtyUI_notify_error(XePUIMain *uimain);

void
a_TtyUI_notify_error_reset(XePUIMain *uimain);

int
a_TtyUI_create_io_watch(XePUI *ui, XePUIIOWatcher *watcher);

void
a_TtyUI_delete_io_watch(XePUI *ui, XePUIIOWatcher *watcher);

#endif
