/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_PYTHON_H

#include <Python.h>

#include "xetpan.h"
#include "ui.h"

#include "xep-python.h"

/* Static pointer to XetPan's main variable to communicate with Python */
static XetPan *global_xetpan;

/* Static functions declarations */

static PyObject *
XePPython_get_folder_list(PyObject *self, PyObject *args);

#if 0
static PyObject *
XePPython_get_mainview_list(PyObject *self, PyObject *args);
#endif

static PyObject *
XePPython_update_folder(PyObject *self, PyObject *args);

/* XetPan Python Module */

static PyMethodDef xetpan_module_def[] =
  {
    {
      "get_folder_list",
      XePPython_get_folder_list,
      METH_VARARGS,
      "Get the name list of all known folders"
    },
#if 0
    {
      "get_mainview_list",
      XePPython_get_mainview_list,
      METH_VARARGS,
      "Get the list of all open mainviews (pointers)"
    },
#endif
    {
      "update_folder",
      XePPython_update_folder,
      METH_VARARGS,
      "Update a folder. Need the reference of a folder as an argument."
    },
    {NULL, NULL, 0, NULL}
  };



int
a_XePPython_is_initialized(void)
{
  return Py_IsInitialized();
}

void
a_XePPython_initialize(XetPan *xetpan)
{
  PyObject *xetpan_module;

  global_xetpan = xetpan;

  Py_Initialize();

  /* Global XetPan variables */
  xetpan_module = Py_InitModule("xetpan", xetpan_module_def);

  PyRun_SimpleString("import xetpan");
}


void
a_XePPython_finalize(void)
{
  Py_Finalize();
}

void
a_XePPython_execute(char *text)
{
  PyRun_SimpleString(text);
}

static PyObject *
XePPython_get_folder_list(PyObject *self, PyObject *args)
{
  PyObject *py_folder_ref;
  PyObject *py_folder_tuple;
  chash *folder_ref_hash_table = a_Folder_ref_hash_get();
  chashiter *iter;
  chashdatum key;
  int i;

  /* Need no arguments */
  if (!PyArg_ParseTuple(args, ""))
    return NULL;

  py_folder_tuple = PyTuple_New(chash_count(folder_ref_hash_table));

  for (i = 0, iter = chash_begin(folder_ref_hash_table); 
       iter;
       iter = chash_next(folder_ref_hash_table, iter), i++)
    {
      chash_key(iter, &key);
      py_folder_ref = PyString_FromString((char *)key.data);
      PyTuple_SetItem(py_folder_tuple, i, py_folder_ref);
    }
  return py_folder_tuple;  
}

#if 0
/* FIXME: mainviews is now a cinthash not a clist */
static PyObject *
XePPython_get_mainview_list(PyObject *self, PyObject *args)
{
  PyObject *py_uimain;
  PyObject *py_uimain_tuple;
  clist *mainviews = global_xetpan->ui->mainviews;
  clistiter *iter;
  int i;

  /* Need no arguments */
  if (!PyArg_ParseTuple(args, ""))
    return NULL;

  py_uimain_tuple = PyTuple_New(clist_count(mainviews));

  for (i = 0, iter = clist_begin(mainviews); 
       iter;
       iter = iter->next, i++)
    {
      py_uimain = PyCObject_FromVoidPtr(iter->data, NULL);
      PyTuple_SetItem(py_uimain_tuple, i, py_uimain);
    }
  return py_uimain_tuple;  
}
#endif
static PyObject *
XePPython_update_folder(PyObject *self, PyObject *args)
{
  const char *folder_ref;

  if (!PyArg_ParseTuple(args, "s", &folder_ref))
    return NULL;

  a_UI_update_folder(folder_ref);

  Py_INCREF(Py_None);
  return Py_None;
}
#endif
